#!/usr/bin/perl
# 
# Copyright (C) 2015 Roland Laifer 
#                    Karlsruhe Institute of Technology (KIT), 
#                    Steinbuch Centre for Computing (SCC)
#                    76128 Karlsruhe, Germany
#                    Roland.Laifer (at) kit.edu
#
# ------------------------------------------------------------------------
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------
#
# $Log: check_quotas.pl,v $
# Revision 1.6  2015/11/16 15:22:22  rz54
# Added check for multiple inode numbers to handle hard links correctly
#
# Revision 1.5  2015/10/29 14:22:26  rz54
# Improved output, fixed wrong order of file percent calculation
#
# Revision 1.4  2015/10/28 16:30:20  rz54
# Added note that the script does not handle hard links correctly
#
# Revision 1.3  2015/10/27 15:41:17  rz54
# Added GPLv3
#
# Revision 1.2  2015/10/27 15:38:22  rz54
# Lots of changes to better conform to good perl coding style
#
# Revision 1.1  2015/10/20 15:46:47  rz54
# Initial revision
#
#
# $Id: check_quotas.pl,v 1.6 2015/11/16 15:22:22 rz54 Exp $

use strict;
use warnings;

use Getopt::Std;
our($opt_D, $opt_d, $opt_h, $opt_t, $opt_v);

# get script name
(my $myname = $0)	=~ s#^.*/##;
# associative array which stores sum of file sizes for entry (user or group) names
undef my %entry2cap;
# associative array which stores number of files for entry names
undef my %entry2fi;
# associative array which stores inode numbers, required to handle hardlinks 
undef my %inodes;
# douser is true if option -u is set, otherwise option -g is set
my $douser = 1;
# entrytype is set to string User or Group and is used to print messsages
my $entrytype; 


sub usage {
    print "Usage: $myname -t u|g -d directory [-h] [-D] [-v]\n";
    print "$myname creates a sum of used capacity by each user or group below a\n";
    print "specified directory (using find and stat commands) and compares the result\n";
    print "with Lustre quotas (using the lfs quota command).\n";
    print " -d directory  directory on which this script operates, usually this is\n";
    print "               the mount point of the Lustre file system\n";
    print " -D            debug\n";
    print " -h            this help\n";
    print " -t            type: user(u) or group (g)\n";
    print " -v            verbose\n\n";
    print "Example: $myname -t g -d /mnt/lustre/\n";
    print "Note that if you run this script during production a difference is of course\n";
    print "normal.\n";
    print "This script was mainly used to check if Lustre quotas are wrong as a result\n";
    print "of the bug reported in https://jira.hpdd.intel.com/browse/LU-4345 .\n";
    print "For details check Roland Laifer's talk at LAD'15,\n";
    print "see http://www.scc.kit.edu/scc/docs/Lustre/kit_lad15_20150922.pdf\n";
    exit 0;
}


sub parse_dir {
    my $cmd;
    if ( $douser ) {
	$cmd = "find $opt_d" . ' -exec stat --printf="%i %u %s\n" \{\} \;';
    } else {
	$cmd = "find $opt_d" . ' -exec stat --printf="%i %g %s\n" \{\} \;';
    }
    print "Executing: $cmd\n" if ($opt_v);
    open my $find_stat_fh, "-|", $cmd or die "*** Command $cmd failed: $!\n";
    while ( my $line = <$find_stat_fh> ) {
#	print "DEBUG: Found line: $line" if ($opt_D);
	if ( $line =~ /^(\d+)\s+(\S+)\s+(\d+)$/ ) {
	    # If inode exists this is a hard link which should not be counted twice. 
	    unless ( $inodes{$1} ) {
		$entry2cap{$2} += $3;
		print "DEBUG: Found: Inode $1, $entrytype $2, size $3\n" if ($opt_D);
		unless ( $entry2fi{$2} ) {
		    $entry2fi{$2} = 1;
		} else {
		    $entry2fi{$2}++;
		}
		$inodes{$1} = 1;
	    } else {
		print "DEBUG: Found: Inode $1 multiple times, i.e. it was not counted.\n" if ($opt_D);
	    }
	} else {
	    print STDERR "Wrong line: $line";
	    print STDERR "Exiting because of unexpected line\n";
	    exit 1;
	}
    }
    close($find_stat_fh);
    if ( $opt_D ) {
	foreach my $entry (sort(keys(%entry2cap))) {
	    print "DEBUG: $entrytype $entry has total size $entry2cap{$entry}\n" if ($opt_D);
	    print "DEBUG: $entrytype $entry has $entry2fi{$entry} files\n" if ($opt_D);
	}
    }
}


sub comp_quotas {
    my ($fsys, $cmd, $quotakb, $quotafi, $gbpercent, $fipercent);
    if ( $opt_d =~ m|^/| ) {
	$fsys = $opt_d;
    } else {
	$fsys = ".";
    }
    foreach my $entry (sort(keys(%entry2cap))) {
	if ( $douser ) {
	    $cmd = "/usr/bin/lfs quota -u $entry $fsys";
	} else {
	    $cmd = "/usr/bin/lfs quota -g $entry $fsys";
	}
	print "Executing: $cmd\n" if ($opt_v);
	my @quotaout = `$cmd`;
	my $match = 0;
	foreach my $line (@quotaout){
	    if ($line =~ /^\s*\S*\s+([0-9]+)[\*]*\s+0\s+[0-9]+\s+[\-\s]*([0-9]+)[\*]*\s+0\s+[0-9]+/) {
		$match = 1;
		$quotakb = "$1";
		$quotafi = "$2";
		print "DEBUG: $entrytype quota of $entry shows $quotakb KB\n" if ($opt_D);
		print "DEBUG: $entrytype quota of $entry shows $quotafi files\n" if ($opt_D);
	    }
	}
	if ($match == 1) {
	    my $quotagb = $quotakb / 1024 / 1024;
	    my $findgb = $entry2cap{$entry} / 1024 / 1024 / 1024;
	    if ($findgb > 0) {
		$gbpercent = $quotagb / $findgb * 100;
	    } else {
		print STDERR "Found capacity for $entrytype $entry has unexpected value ($findgb).";
		$gbpercent = 0;
	    }
	    if ($entry2fi{$entry} > 0) {
		$fipercent = $quotafi / $entry2fi{$entry} * 100;
            } else {
                print STDERR "Found files for $entrytype $entry has unexpected value ($entry2fi{$entry}).";
		$fipercent = 0;
	    }
	    my $name = "NOTFOUND";
	    if ( $douser ) {
		$cmd = "/usr/bin/getent passwd $entry";
	    } else {
		$cmd = "/usr/bin/getent group $entry";
	    }
	    print "Executing: $cmd\n" if ($opt_v);
	    my @getentout = `$cmd`;
	    foreach my $line (@getentout){
		if ($line =~ /^([^:]+):/) {
		    $name = "$1";
		}
	    }
	    printf("$entrytype %4s (ID %4d) has %5.2f GB quotas and %5.2f GB total size (%2.1f percent)\n", $name, $entry, $quotagb, $findgb, $gbpercent);
	    printf("$entrytype %4s (ID %4d) has %5d quota files and %5d found files (%2.1f percent)\n", $name, $entry, $quotafi, $entry2fi{$entry}, $fipercent);
	} else {
	    print STDERR "Error parsing output of command $cmd\nOutput: @quotaout";
	}
    }
}


# get options
getopts("Dd:ht:v") || usage;
usage if ( $opt_h );
usage unless ( $opt_d );
usage unless ( $opt_t );
usage unless ($#ARGV == -1);

if ( $opt_t =~ /^u$/ ) {
    $douser = 1;
    $entrytype = "User";
} elsif ( $opt_t =~ /^g$/ ) {
    $douser = 0;
    $entrytype = "Group";
} else {
    print STDERR "Either option -tu or option -tg is required!\n";
    usage;
}

parse_dir;

comp_quotas;

exit 0;
